/*
 * Copyright (C) 2008-2009 Alex Lasnier <alex@fepg.org>
 *
 * This file is part of NAScan
 *
 * NAScan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * NAScan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NAScan.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <stdarg.h>

#include <vdr/plugin.h>

#include "osd.h"
#include "tools.h"
#include "config.h"


#define SLEEP_TIME  100

#define SOURCES_FILE "sources.conf"


//////////////////////////////////////////////////////////////////////////////


ScanOsd::ScanOsd(void) : cOsdMenu("NA Channel Scanner"), cThread("NA Scan"), nitFilter(&transponderList), pmtFilter(&channelList)
{
  SetHelp("Cancel");
	
	dir = cPlugin::ConfigDirectory("nascan");

	numSources  = 0;
	needsUpdate = 0;
	
  Display();
  Start(); 
}


//----------------------------------------------------------------------------

ScanOsd::~ScanOsd(void) 
{
  Cancel(SLEEP_TIME+500);
}


//----------------------------------------------------------------------------

void ScanOsd::Action(void)
{
  if (!ReadSources())
    return;
  
  time_t startTime = time(NULL);
  
  // Start scanning the sources
  for (int i=0; i<numSources && Running(); i++)
  {
    AddLine("--- Scanning %s %s -------------------------------------------------------------", 
            *cSource::ToString(sources[i]), Sources.Get(sources[i])->Description());
    
    // Tune initial transponder: get SDTs and NITs 
    cChannel* initial = GetInitial(i);

    cDevice* device = NULL;
    if (config.adapter == 0) // Auto
    {
#if VDRVERSNUM < 10500  
      device = cDevice::GetDevice(initial);
#else
      device = cDevice::GetDevice(initial, -1, false);
#endif  
    }
    else device = cDevice::GetDevice(config.adapter - 1);
    
    if (!device) {
      AddLine("ERROR: No device found.");
      return;
    }
    
    if (!Tune(device, initial, config.initialTuningTimeout)) {
      AddLine("ERROR: Unable to tune initial transponder.");
      continue;
    }
    
    delete initial; initial = NULL;
    
    AddLine("Downloading all SDTs/NITs, this may take a while...");
    
    sdtFilter.Attach(device);
    nitFilter.Attach(device);
    
    cCondWait::SleepMs(1000);
    
    int timeout = 0;
    while (Running() && !nitFilter.GotNit() && timeout <= config.nitTimeout) {
      cCondWait::SleepMs(SLEEP_TIME);
      if (config.nitTimeout && !nitFilter.GotData()) 
        timeout += SLEEP_TIME;
    }
    if (Running() && nitFilter.GotNit())
      AddLine("Received all NITs: found %d transponders", transponderList.Count());
    
    if (nitFilter.GotNit())
      while (Running() && !sdtFilter.GotSdts() && timeout <= config.sdtTimeout) {
        cCondWait::SleepMs(SLEEP_TIME);
        if (config.sdtTimeout && !sdtFilter.GotData()) 
          timeout += SLEEP_TIME;
      }
    sdtFilter.Detach();
    nitFilter.Detach();
    
    if (!Running()) break;
    
    if (!sdtFilter.GotSdts() || !nitFilter.GotNit()) {
      AddLine("ERROR: SDT/NIT Timeout (%d %d).", sdtFilter.GotSdts(), nitFilter.GotNit());
      AddLine("");
      continue;    
    }
    
    AddLine("Received all SDTs");

    
    // Scan each transponder
    transponderList.Sort();
    pmtFilter.Attach(device);
    
    for (TransponderSorter* cs = transponderList.First() ; cs && Running(); cs = transponderList.Next(cs)) 
    {
      pmtFilter.SetTransponder(cs->channel);
      
      if (!Tune(device, cs->channel, config.tuningTimeout))
        continue;
      
      cCondWait::SleepMs(1000);
      
      // Wait for PMTs  
      timeout = 0;
      while (Running() && !pmtFilter.GotPMT() && timeout <= config.pmtTimeout) {
        cCondWait::SleepMs(SLEEP_TIME);
        if (config.pmtTimeout)
          timeout += SLEEP_TIME;
      }      
      if (!Running()) break;
      
      if (!pmtFilter.GotPMT()) {
        AddLine("ERROR: PMT Timeout. (%d)", pmtFilter.GotPAT());
        continue;
      }
      
      AddLine("  Found %d channels", pmtFilter.GotPMT());
    }
    if (!Running()) break;
    
    pmtFilter.Detach();

    //channelList.Sort();
    int numChannels = 0;
    
    FILE* outFile = NULL; 
    char* fn = NULL;
    asprintf(&fn, "%s/channels-%s.conf", dir, *cSource::ToString(sources[i]));
  
    outFile = fopen(fn, "w");
    free(fn); fn = NULL;
    if (outFile) {
      for (ChannelSorter* cs = channelList.First(); cs; cs = channelList.Next(cs)) {
        cChannel* c = cs->channel;
        
        const char* name = names.Name(c->Nid(), c->Tid(), c->Sid());
        if (!name)
          name = sdtFilter.GetName(c->Nid(), c->Tid(), c->Sid());
        
        const char* shortName = sdtFilter.GetShortName(c->Nid(), c->Tid(), c->Sid());
        const char* networkName = config.useNetworkName ? nitFilter.GetNetworkName(c->Nid()) : "";
        
        if (name && shortName)
        {
          c->SetName(name, shortName, networkName);
          cString chanText = c->ToText();
          fprintf(outFile, "%s", *chanText);
          numChannels++;
        }
      }
      fclose(outFile);
    }
    
    // TEMP --- START ---
/*
    asprintf(&fn, "%s/names-%s.conf", dir, *cSource::ToString(sources[i]));
    outFile = fopen(fn, "w");
    free(fn); fn = NULL;
    if (outFile) {
      for (ChannelSorter* cs = channelList.First(); cs; cs = channelList.Next(cs)) {
        cChannel* c = cs->channel;
        const char* name = sdtFilter.GetName(c->Nid(), c->Tid(), c->Sid());
        if (name) 
        {
          fprintf(outFile, "%s:%d:%d:%d\n", name , c->Sid(), c->Nid(), c->Tid());
        }
      }
      fclose(outFile);
    }  
*/      
    // TEMP --- END ---
    
    
    channelList.Clear();
    
    transponderList.Clear();
    
    AddLine("Found a total of %d channels on %s", numChannels, Sources.Get(sources[i])->Description());
    AddLine("");
  }
  
  pmtFilter.Detach();
  if (Running())
    AddLine("Done (took %d seconds)", time(NULL) - startTime);
}


//----------------------------------------------------------------------------

eOSState ScanOsd::ProcessKey(eKeys Key)
{
  eOSState state = cOsdMenu::ProcessKey(Key);

  if (state == osUnknown)
  {
    state = osContinue;
    
    switch (Key)
    {
      case kBack:
        state = osBack;
      break;
      
      case kRed:
        if (Running()) {
          Cancel(SLEEP_TIME+500);
          AddLine("Scan cancelled");
        }
        else
          state = osBack;
      break;
      
      case kNone:
        if (needsUpdate) {
          Display();
          needsUpdate = false;
        }
      break;
      
      default:
      break;    
    }
  }  
  return state;  
}


//----------------------------------------------------------------------------

void ScanOsd::AddLine(const char* text, ...) 
{
  char* buffer = NULL;
  va_list ap;
  va_start(ap, text);
  vasprintf(&buffer, text, ap);
  va_end(ap);

  cOsdItem* item = new cOsdItem(buffer);
  dprint(L_MSG, "%s", buffer);
  free(buffer);

  cOsdMenu::Add(item);
  CursorDown();
  needsUpdate = true;
}


//----------------------------------------------------------------------------

void ScanOsd::UpdateLastLine(const char* text)
{
  char* buffer = NULL;
  asprintf(&buffer, "%s %s", Last()->Text(), text);
  Last()->SetText(buffer, false); // false, so we don't need to free(buffer)
  needsUpdate = true;
  dprint(L_MSG, "%s", text);
}


//----------------------------------------------------------------------------

bool ScanOsd::ReadSources(void)
{
  char* fn = NULL;
  FILE* file = NULL;
  
  asprintf(&fn, "%s/%s", dir, SOURCES_FILE);
  
  file = fopen(fn, "r");
  if (file == NULL) { 
    AddLine("ERROR: Could not open sources file:");
    AddLine(fn);
    free(fn);
    return false;
  }
  free(fn);
  
  char* s;
  int line = 0;
  cReadLine ReadLine;
  bool result = true;
  
  while ((s = ReadLine.Read(file)) != NULL) 
  {
    line++;
    char *p = strchr(s, '#');
    if (p) *p = 0;
    stripspace(s);
    if (!isempty(s)) {
      char* src = NULL;
      int freq = 0;
      char pol = 0;
      int sr = 0;
      
      bool ok = ( sscanf(s, "%a[^:]:%d :%c :%d ", &src, &freq ,&pol ,&sr) == 4);
      int source = cSource::FromString(src);
      free(src);
      
      pol = toupper(pol);
      
      if (ok && source && freq && sr && (pol == 'H' || pol == 'V')) {
        if (numSources < MAXSOURCES) {
          sources[numSources]      = source;
          frequency[numSources]    = freq;
          polarization[numSources] = pol;
          symbolRate[numSources]   = sr;
          numSources++;
        }
      }
      else {
        AddLine("ERROR: error in %s, line %d", SOURCES_FILE, line); 
        result = false;
        break;
      }
    }
  }
  fclose(file);
  
  return result;
}


//----------------------------------------------------------------------------

cChannel* ScanOsd::GetInitial(int i) const
{
  cChannel* c = new cChannel();

  int cr  = MapToDriver(999, CoderateValues);   // Auto
  
#if VDRVERSNUM < 10700
  c->SetSatTransponderData(sources[i], frequency[i], polarization[i], symbolRate[i], cr);
#else
  int mod = MapToDriver(999, ModulationValues); // Auto 
  int sys = MapToDriver(0, SystemValues);       // DVB-S
  int ro  = MapToDriver(0, RollOffValues);      // Unknown
  
  c->SetSatTransponderData(sources[i], frequency[i], polarization[i], symbolRate[i], cr, mod, sys, ro);
#endif

  return c; 
}


//----------------------------------------------------------------------------

bool ScanOsd::Tune(cDevice* device, const cChannel* c, int timeout)
{
  AddLine("Tuning %d MHz ...", c->Frequency());

  if (!device->SwitchChannel(c, false)) {
    UpdateLastLine("Failed");
    return false;  
  }
  
  bool lock = device->HasLock(timeout); 
  UpdateLastLine(lock ? "Success" : "Failed (Timeout)");

  return lock;
}


//////////////////////////////////////////////////////////////////////////////


