/*
 * Copyright (C) 2008-2009 Alex Lasnier <alex@fepg.org>
 *
 * This file is part of NAScan
 *
 * NAScan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * NAScan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NAScan.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ctype.h>

#include <vdr/channels.h>

#include <libsi/section.h>
#include <libsi/descriptor.h>

#include "pmtfilter.h"
#include "tools.h"
#include "config.h"


//////////////////////////////////////////////////////////////////////////////


#define PAT_PID    0x0000

#define PAT_TID    0x00
#define PMT_TID    0x02


//----------------------------------------------------------------------------

PmtFilter::PmtFilter(cList<ChannelSorter>* cl)
{
  channelList = cl;

  numPmtSids = 0;
  pmtSidIndex = 0;
  source = transponder = 0;
  
  gotPat = false;
  gotPmts = 0;
    
  curTransponder = NULL;
  
  Set(PAT_PID, PAT_TID);
}


//----------------------------------------------------------------------------

PmtFilter::~PmtFilter()
{

}


//----------------------------------------------------------------------------

void PmtFilter::SetStatus(bool On)
{   
  if (On && source == Source() && transponder == Transponder())
  {
    dprint(L_DBG, "Status change: same source, same transponder. No reset.");
  }
  else if (On) 
  {
    dprint(L_DBG, "Status change: resetting scanner");
    
    gotPat = false;
    gotPmts = 0;
    
    pmtSids.clear();
    numPmtSids = 0;
    pmtSidIndex = 0;

    source = Source();
    transponder = Transponder();
  }
 
  cFilter::SetStatus(On);
}


//----------------------------------------------------------------------------

void PmtFilter::Process(u_short pid, u_char tid, const u_char* data, int length)
{ 
  switch (tid)
  {
    case PAT_TID:
      if (!gotPat && pid == PAT_PID)
        if (ProcessPAT(data)) {
          gotPat = true;
        }
    break;
     
    case PMT_TID:
      if (curTransponder)
        if (ProcessPMT(data)) 
        {
          Del(pid, tid);
          
          if (pmtSids.empty()) {
            dprint(L_PMT, "Received all PMTs.");
            gotPmts = numPmtSids;
          }
        }
    break;
    
    default:
      dprint(L_ERR, "PmtFilter: Unknown TID 0x%02X.", tid);
    break;   
  }
}


//----------------------------------------------------------------------------

bool PmtFilter::ProcessPAT(const uint8_t* data)
{
  SI::PAT pat(data, false);
  if (!pat.CheckCRCAndParse())
    return false;
    
  dprint(L_PAT, "Received PAT.");
  
  SI::PAT::Association assoc;
  for (SI::Loop::Iterator it; pat.associationLoop.getNext(assoc, it); ) 
  {
    if (!assoc.isNITPid()) {
      dprint(L_PAT, "PAT: Found PMT (pid: %d, sid: %d)", assoc.getPid(), assoc.getServiceId());
      Add(assoc.getPid(), PMT_TID);
      pmtSids.push_back(assoc.getServiceId());
    }  
  }
  numPmtSids = pmtSids.size();
  return true;
}


//----------------------------------------------------------------------------

template <class T> bool Included(std::list<T>& list, T val)
{
  typename std::list<T>::iterator itr;
  for(itr = list.begin(); itr != list.end() && val != *itr; itr++) { /* do nothing */ }
  if (itr == list.end()) // not found
    return false;

  list.erase(itr); // If the item is found, delete it.
  return true;
}


//----------------------------------------------------------------------------

bool PmtFilter::ProcessPMT(const uint8_t* data)
{
  SI::PMT pmt(data, false);
  if (!pmt.CheckCRCAndParse())
    return false;
  
  if (!Included(pmtSids, (uint16_t) pmt.getServiceId())) // This is probably unnecessary...
  {
    dprint(L_PMT, "Received PMT sid: %d (Already seen).", pmt.getServiceId());
    return false;
  }

  dprint(L_PMT, "Received PMT sid: %d (%d of %d).", pmt.getServiceId(), ++pmtSidIndex, numPmtSids);
  
  if (config.ignoreNoStream && pmt.streamLoop.getLength() == 0) {
    dprint(L_PMT, "PMT: no stream info, ignoring.");
    return true;
  }

  if (blacklist.Blacklisted(curTransponder->Nid() , curTransponder->Tid(), pmt.getServiceId())) {
    dprint(L_PMT, "PMT: channel is blacklisted, ignoring.", pmt.getServiceId());
    return true;    
  }

  int NumCaIds = 0;
  int CaIds[MAXCAIDS + 1] = { 0 };
  
  /*SI::CaDescriptor* d;
  for (SI::Loop::Iterator it; (d = (SI::CaDescriptor*)pmt.commonDescriptors.getNext(it, SI::CaDescriptorTag)); ) {
    //dprint(L_PMT, "CAID: 0x%04X", d->getCaType());
    if (NumCaIds < MAXCAIDS) 
      CaIds[NumCaIds++] = d->getCaType();      
    delete d;
  } */
  
  int Vpid = 0;
  int Ppid = pmt.getPCRPid();
  int Apids[MAXAPIDS + 1] = { 0 };
  int Dpids[MAXDPIDS + 1] = { 0 };
  int Spids[MAXSPIDS + 1] = { 0 };
  char ALangs[MAXAPIDS][MAXLANGCODE2] = { "" };
  char DLangs[MAXDPIDS][MAXLANGCODE2] = { "" };
  char SLangs[MAXSPIDS][MAXLANGCODE2] = { "" };
  int Tpid = 0;
  int NumApids = 0;
  int NumDpids = 0;
  // int NumSpids = 0; // Don't need this since we don't care about Spids
  
  SI::PMT::Stream stream;
   
  for (SI::Loop::Iterator it; pmt.streamLoop.getNext(stream, it); ) 
  {
    switch (stream.getStreamType())
    {
      case 0x02: // Video
      //case 0x1B: // MPEG-4 ?
        Vpid = stream.getPid();
      break;
      
      case 0x04: // STREAMTYPE_13818_AUDIO
      {
        if (NumApids < MAXAPIDS) {
          Apids[NumApids] = stream.getPid();
          SI::Descriptor *d;
          for (SI::Loop::Iterator it; (d = stream.streamDescriptors.getNext(it)); ) {
            switch (d->getDescriptorTag()) {
              case SI::ISO639LanguageDescriptorTag: {
                SI::ISO639LanguageDescriptor *ld = (SI::ISO639LanguageDescriptor *)d;
                SI::ISO639LanguageDescriptor::Language l;
                char *s = ALangs[NumApids];
                int n = 0;
                for (SI::Loop::Iterator it; ld->languageLoop.getNext(l, it); ) {
                  if (*ld->languageCode != '-') { 
                    if (n > 0)
                      *s++ = '+';
                    strn0cpy(s, I18nNormalizeLanguageCode(l.languageCode), MAXLANGCODE1);
                    s += strlen(s);
                    if (n++ > 1)
                      break;
                  }
                }
              }
              break;
              default: ;
            }
            delete d;
          }
          NumApids++;
        }
      }      
      break;
      
      case 0x81: // AC-3 audio
      {
        if (NumDpids < MAXDPIDS) {
          Dpids[NumDpids] = stream.getPid();
          SI::Descriptor *d;
          for (SI::Loop::Iterator it; (d = stream.streamDescriptors.getNext(it)); ) {
            switch (d->getDescriptorTag()) {
              case SI::ISO639LanguageDescriptorTag: {
                SI::ISO639LanguageDescriptor *ld = (SI::ISO639LanguageDescriptor *)d;
                SI::ISO639LanguageDescriptor::Language l;
                char *s = DLangs[NumDpids];
                int n = 0;
                for (SI::Loop::Iterator it; ld->languageLoop.getNext(l, it); ) {
                  if (*ld->languageCode != '-') { 
                    if (n > 0)
                      *s++ = '+';
                    strn0cpy(s, I18nNormalizeLanguageCode(l.languageCode), MAXLANGCODE1);
                    s += strlen(s);
                    if (n++ > 1)
                      break;
                  }
                }
              }
              break;
              default: ;
            }
            delete d;
          }
          NumDpids++;
        }
      }  
      break;
      
      default:
        dprint(L_PMT, "PMT: Unknown stream type 0x%02X.", stream.getStreamType());
      break;
    }
    
    SI::CaDescriptor* d;
    for (SI::Loop::Iterator it; (d = (SI::CaDescriptor*) stream.streamDescriptors.getNext(it, SI::CaDescriptorTag)); ) 
    {
      if (NumCaIds < MAXCAIDS) {
        int caId = d->getCaType();
        bool add = true;
        for (int i = 0; add && i < NumCaIds; i++)
          if (CaIds[i] == caId)
            add = false;
        if (add)    
          CaIds[NumCaIds++] = caId;
      }        
      delete d;
    }
  }
  
  cChannel* channel = new cChannel();
  channel->CopyTransponderData(curTransponder);
  channel->SetId(curTransponder->Nid() , curTransponder->Tid(), pmt.getServiceId());
  channel->SetPids(Vpid, Vpid ? Ppid : 0, Apids, ALangs, Dpids, DLangs, Spids, SLangs, Tpid);
  channel->SetCaIds(CaIds);
  
  //channel->SetNumber(pmt.getServiceId());
  
  channelList->Add(new ChannelSorter(channel));
  
  return true;    
}


//////////////////////////////////////////////////////////////////////////////


