/*
 * Copyright (C) 2008-2009 Alex Lasnier <alex@fepg.org>
 *
 * This file is part of NAScan
 *
 * NAScan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * NAScan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NAScan.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libsi/section.h>
#include <libsi/descriptor.h>

#include "sdtfilter.h"
#include "tools.h"


#define SDT_KEY(NID,TSID,SID) ((((uint64_t) NID) << 32) | (((uint64_t) TSID) << 16) | ((uint64_t) SID))

#define SDT_PID    0x0011
#define SDT_TID    0x42  // Actual transport stream
#define SDT_TID_O  0x46  // Other transport stream 


//////////////////////////////////////////////////////////////////////////////


SdtFilter::SdtFilter(void)
{
  gotAllSdts = gotActualSdt = gotSdtData = false;
  source = 0;
  
  Set(SDT_PID, SDT_TID);
  Set(SDT_PID, SDT_TID_O);
}


//----------------------------------------------------------------------------

SdtFilter::~SdtFilter(void)
{

}


//----------------------------------------------------------------------------

const char* SdtFilter::GetShortName(uint16_t nid, uint16_t tid, uint16_t sid) const
{
  std::map<uint64_t, ChannelName>::const_iterator itr = channelNames.find(SDT_KEY(nid, tid, sid));
  if (itr != channelNames.end())
    return (*itr).second.shortName.c_str();
  else {
    dprint(L_SDT, "No name for %d %d %d", nid, tid, sid);
    return NULL;   
  }
}


//----------------------------------------------------------------------------

const char* SdtFilter::GetName(uint16_t nid, uint16_t tid, uint16_t sid) const
{
  std::map<uint64_t, ChannelName>::const_iterator itr = channelNames.find(SDT_KEY(nid, tid, sid));
  if (itr != channelNames.end())
    return (*itr).second.name.c_str();
  else {
    return NULL;   
  }
}

//----------------------------------------------------------------------------

void SdtFilter::SetStatus(bool On)
{
  if (On && gotAllSdts && source == Source())
  {
    dprint(L_DBG, "SDT Filter: same source. No reset.");
  }
  else if (On)
  {
    dprint(L_DBG, "SDT Filter: diffrent source. Clearing SDT cache.");
    gotAllSdts = gotActualSdt = false;
    gotSdtData = false;
    channelNames.clear();
    source = Source();
  }
  
  cFilter::SetStatus(On);
}


//----------------------------------------------------------------------------

void SdtFilter::Process(u_short pid, u_char tid, const u_char* data, int length)
{ 
  if (tid == SDT_TID) { // Actual TS SDT may cause premature duplicate detection
    if (gotActualSdt) {
      dprint(L_DBG, "SDT: Actual TS SDT already seen."); 
      return;
    }  
    gotActualSdt = true;
  }
  
  SI::SDT sdt(data, false);
  if (!sdt.CheckCRCAndParse())
    return;
  
  gotSdtData = true;
 
  SI::SDT::Service SiSdtService;
  for (SI::Loop::Iterator it; sdt.serviceLoop.getNext(SiSdtService, it); )
  {
    uint64_t key = SDT_KEY(sdt.getOriginalNetworkId(), sdt.getTransportStreamId(), SiSdtService.getServiceId());
    
    std::map<uint64_t, ChannelName>::const_iterator itr = channelNames.find(key);
    if (itr != channelNames.end()) { // Key found
      // When we start getting duplicates, it's probably safe to stop.
      dprint(L_SDT, "SDT: Received duplicate SDT (%d %d %d)", sdt.getOriginalNetworkId(), sdt.getTransportStreamId(), SiSdtService.getServiceId());
      gotAllSdts = true;
      SetStatus(false);
      return;
    }
    
    SI::Descriptor* d;
    for (SI::Loop::Iterator it2; (d = SiSdtService.serviceDescriptors.getNext(it2)); ) 
    {
      switch (d->getDescriptorTag()) 
      {
        case SI::ServiceDescriptorTag: // 0x48
        {
          SI::ServiceDescriptor *sd = (SI::ServiceDescriptor*) d;
          dprint(L_SDT, "SDT: Service Descriptor type: 0x%02X", sd->getServiceType());
          
          ChannelName cn;
          
          char nameBuf[sd->serviceName.getLength()+1];
          sd->serviceName.getText(nameBuf, sizeof(nameBuf));
          
          dprint(L_SDT, "     %s", nameBuf);
          
          cn.shortName = !isempty(nameBuf) ? nameBuf : "No Name";

          char providerNameBuf[sd->providerName.getLength()+1];

          //if (!isempty(providerNameBuf) && strcmp(nameBuf, providerNameBuf)) dprint(L_MSG, "{0x%02X} %s --- %s",sd->getServiceType() , nameBuf, providerNameBuf);

          // TODO: Ignore certain channels based on service type.
        
          switch (sd->getServiceType())
          {
            case 0x01: // digital television service
            case 0x02: // digital radio sound service
            case 0x80: // WEBTV Download (could ignore)
            case 0x85: // ?
            case 0x88: // ?
            case 0x8A: // ? 
            case 0x8D: // EPG (has useful name)
            case 0x96: // ?
            case 0xA0: // ?
            //
            case 0xA4: // DN HD ?
            case 0xA8: // ?
            case 0xFF: // ?
              sd->providerName.getText(providerNameBuf, sizeof(providerNameBuf));
            break;
            
            case 0x9A: // ?
              sd->providerName.getText(providerNameBuf, sizeof(providerNameBuf));
              if (strcmp(providerNameBuf, "SMART") == 0) // probably temp.
                providerNameBuf[0] = 0;
            break;

            case 0x86: // ?
            case 0x89: // Interactive ? 
            case 0x8C: // ?
            case 0x8F: // ?
            case 0x90: // ?
            case 0x92: // ?
            case 0x9B: // ?
            case 0xA1: // ?
            case 0xA2: // VOD ?
            case 0xA3: // VOD ?
            case 0xA5: // ?
            case 0xA7: // ?
            case 0xAD: // ?
            case 0xAE: // ?
            case 0xB0: // ?
              providerNameBuf[0] = 0;
            break;
              
            default:
              sd->providerName.getText(providerNameBuf, sizeof(providerNameBuf));
              dprint(L_ERR, "SDT:Unkown service type 0x%02X | %s --- %s",sd->getServiceType() , nameBuf, providerNameBuf);
              providerNameBuf[0] = 0;
            break;
          }
          
          cn.name = !isempty(providerNameBuf) ? (!isempty(nameBuf) ? providerNameBuf : "No Name") : nameBuf; 
          channelNames[key] = cn;         
        }
        break;
        
        case SI::LinkageDescriptorTag: // 0x4A: EPG Link
          dprint(L_SDT, "SDT: Linkage Descriptor");
        break;

        case 0x80: // Dish channel mapping
        case 0x82: // EPG channel
        case 0x84: // ???  
        case 0x85: // Password/Region protection     
        case 0x86: // Tier descriptor     
        case 0x93: // Tier rights descriptor
        case 0x9E: // Dish channel available
        case 0x99: // ATSC ?
          dprint(L_SDT, "SDT: Unsupported tag: 0x%02X", d->getDescriptorTag());
        break;
        
        default:
          dprint(L_SDT, "SDT: Unknown tag: 0x%02X", d->getDescriptorTag());
        break;
      }
      
      delete d;
      
    }
  }
}


//////////////////////////////////////////////////////////////////////////////


